import numpy as np
import vg
from .._common.validation import check_arity, check_indices


def reconcile_selection(faces, face_mask, vertex_mask, prune_orphan_vertices):
    """
    Reconcile the given vertex and face masks. When vertices are removed,
    their faces must also be removed. When faces are removed, the vertices
    can be removed or kept, depending on the `prune_orphan_vertices` option.

    Args:
        faces (np.ndarray): A `kx3` or `kx4` array of the vertices of each
            face.
        face_mask (np.ndarray): A boolean face mask.
        vertex_mask (np.ndarray): A boolean vertex mask.
        prune_orphan_vertices (bool): When `True`, remove vertices which
            their last referencing face is removed.

    Returns:
        tuple: The reconciled vertex and face masks.
    """
    num_faces, _ = vg.shape.check(locals(), "faces", (-1, -1))
    check_arity(faces)
    vg.shape.check(locals(), "face_mask", (num_faces,))
    num_vertices = vg.shape.check(locals(), "vertex_mask", (-1,))
    if face_mask.dtype != np.bool or vertex_mask.dtype != np.bool:
        raise ValueError("Expected face_mask and vertex_mask to be boolean arrays")
    check_indices(faces, num_vertices, "faces")

    # Invalidate faces containing any vertex which is being removed.
    reconciled_face_mask = np.zeros_like(face_mask, dtype=np.bool)
    reconciled_face_mask[face_mask] = np.all(vertex_mask[faces[face_mask]], axis=1)

    # Optionally, invalidate vertices for faces which are being removed.
    if prune_orphan_vertices:
        # Orphaned verts are those belonging to faces which are being
        # removed, and not faces which are being kept.
        orphaned_vertices = np.setdiff1d(
            faces[~reconciled_face_mask], faces[reconciled_face_mask]
        )
        reconciled_vertex_mask = np.copy(vertex_mask)
        reconciled_vertex_mask[orphaned_vertices] = False
    else:
        reconciled_vertex_mask = vertex_mask

    return reconciled_face_mask, reconciled_vertex_mask



import numpy as np
import vg


class GroupMap:
    """
    An immutable map of groups of elements, which are allowed to overlap.
    These can be used for face or vertex groups, as in the Wavefront OBJ
    standard.

    Args:
        num_elements (int): The total number of elements. This determines
            the length of the masks.
        group_names (list): The names of the groups.
        masks (np.array): A boolean array with a row containing a boolean
            mask for each group.

    See also:
        http://paulbourke.net/dataformats/obj/
    """

    def __init__(self, num_elements, group_names, masks, copy_masks=False):
        if not isinstance(num_elements, int) or num_elements < 0:
            raise ValueError("num_elements should be a non-negative integer")
        if not all(isinstance(group_name, str) for group_name in group_names):
            raise ValueError("group_names should be a list of strings")
        vg.shape.check(locals(), "masks", (len(group_names), num_elements))
        if masks.dtype != np.bool:
            raise ValueError("Expected masks to be a bool array")

        if copy_masks:
            masks = masks.copy()
        masks.setflags(write=False)
        self._num_elements = num_elements
        self._masks = masks
        self._group_names = {k: i for i, k in enumerate(group_names)}

    @classmethod
    def from_dict(cls, group_data, num_elements):
        """
        Create a group map from a dictionary of elements. The keys are the
        group names and the values are lists of element indices.

        Args:
            group_data (dict): The group data.
            num_elements (int): The total number of elements.
        """
        masks = np.zeros((len(group_data), num_elements), dtype=np.bool)
        for i, element_indices in enumerate(group_data.values()):
            try:
                masks[i][element_indices] = True
            except IndexError:
                raise ValueError(
                    "Element indices should be less than {}".format(num_elements)
                )
        return cls(
            num_elements=num_elements,
            group_names=group_data.keys(),
            masks=masks,
            copy_masks=False,
        )

    def __len__(self):
        """
        Get the number of groups.

        Returns:
            int: The number of groups.
        """
        return len(self._group_names)

    def __iter__(self):
        """
        Iterate over the groups.

        Returns:
            list_iterator: An iterator over the groups.
        """
        return iter(self._group_names)

    def __getitem__(self, group_name):
        """
        Get the read-only mask for the requested group.

        Args:
            group_name (string): The desired group.

        Returns:
            np.array: A read-only boolean array with length equal to
            `self.num_elements`.
        """
        try:
            index = self._group_names[group_name]
        except KeyError:
            raise KeyError("Unknown group: {}".format(group_name))
        return self._masks[index]

    def keys(self):
        """
        Get the names of all the groups.

        Returns:
            list: A list of the group names.
        """
        return list(self._group_names)

    @property
    def num_elements(self):
        return self._num_elements

    def mask_for_element(self, element):
        """
        Get the read-only group mask for the requested element.

        Args:
            element (int): The desired element.

        Returns:
            np.array: A read-only boolean array corresponding to the
                group names in `self.keys()`.
        """
        return self._masks[:, element]

    def group_names_for_element_mask(self, element_mask):
        """
        Translate an element mask to a list of group names.

        Args:
            element_mask (np.array): An element mask (e.g. a return value from
                `mask_for_element()`).

        Returns:
            list: The group membership represented by the element mask.
        """
        vg.shape.check(locals(), "element_mask", (len(self._group_names),))
        group_names = self.keys()
        return [group_names[index] for index in element_mask.nonzero()[0]]

    def union(self, *group_names):
        """
        Construct the union of the requested groups and return it as a
        writable mask.

        Args:
            group_names (list): The requested groups.

        Returns:
            np.array: A boolean mask with length equal to `self.num_elements`.
        """
        if not all(isinstance(group_name, str) for group_name in group_names):
            raise ValueError("Group names must be strings")
        indices = []
        invalid_group_names = []
        for group_name in group_names:
            try:
                indices.append(self._group_names[group_name])
            except KeyError:
                invalid_group_names.append(group_name)
        if len(invalid_group_names):
            raise KeyError("Unknown groups: {}".format(", ".join(invalid_group_names)))
        return np.any(self._masks[indices], axis=0)

    def reindexed(self, f_new_to_old):
        """
        Given a mapping from new face indices to old face indices, construct
        a group map which preserves the original groups but references the
        new set of indices. When reindexing a mesh, invoke this function on
        the old group map to construct a new group map which preserves the
        original segments wherever possible.

        Args:
            f_new_to_old (np.ndarray): The old face index
                corresponding to each of the new faces.

        Returns:
            GroupMap: A new group map suitable for use with the new faces.
        """
        num_elements = len(f_new_to_old)
        # new_masks =np.zeros((len(self), num_elements), dtype=np.bool)
        # new_masks[f_new_to_old] = self._masks
        new_masks = np.asarray(self._masks[:, f_new_to_old])
        return GroupMap(
            num_elements=num_elements,
            group_names=self._group_names,
            masks=new_masks,
            copy_masks=False,
        )



from collections import OrderedDict
import numpy as np
from .._group_map import GroupMap
from .._mesh import Mesh

try:
    from tinyobjloader import ObjReader, ObjReaderConfig
except Exception:  # pragma: no cover
    ObjReader = None
    ObjReaderConfig = None


class LoadException(Exception):
    pass


class ArityException(Exception):
    pass


def unstack(stacked_result, slice_lengths, safe=True):
    """
    Iterate over slices within a stacked result.
    """
    if safe:
        k = np.sum(slice_lengths)
        if stacked_result.shape[0] != k:  # pragma: no cover
            raise ValueError("Expected stacked_result to have lenth {}".format(k))

    # This code is equivalent to this NumPy call. Strangely enough, our
    # generator-based implementation seems to be faster.
    # return np.array_split(stacked_result, np.cumsum(slice_lengths)[:-1], axis=0)

    last_index = -1
    for slice_length in slice_lengths:
        new_last_index = last_index + int(slice_length)
        yield stacked_result[last_index + 1 : new_last_index + 1]
        last_index = new_last_index


def create_reader_and_config():
    if ObjReader is None:  # pragma: no cover
        raise ImportError(
            """
            tinyobjloader library has not been installed.
            You will not be able to load OBJ files.
            To fix, run `pip install lacecore[obj]`.
            """
        )
    reader = ObjReader()
    config = ObjReaderConfig()
    # There is some complex code in tinyobjloader which occasionally switches
    # the axes of triangulation based on the vertex positions. This is
    # undesirable in lacecore as it scrambles correspondence.
    config.triangulate = False
    return reader, config


def _finalize(reader, triangulate):
    attrib = reader.GetAttrib()
    shapes = reader.GetShapes()
    tinyobj_vertices = attrib.numpy_vertices().reshape(-1, 3)
    if len(shapes) > 0:
        all_vertices_per_face = np.concatenate(
            [shape.mesh.numpy_num_face_vertices() for shape in shapes]
        )
        first_arity = all_vertices_per_face[0]
        if np.any(all_vertices_per_face < 3) or np.any(all_vertices_per_face > 4):
            raise ArityException(
                "OBJ Loader does not support arities greater than 4 or less than 3"
            )
        is_mixed_arity = np.any(all_vertices_per_face != first_arity)
        index_dtype = all_vertices_per_face.dtype
        all_faces = np.zeros((0, 3 if triangulate else first_arity), dtype=index_dtype)
    else:
        index_dtype = np.int32
        all_faces = np.zeros((0, 3), dtype=index_dtype)

    segm = OrderedDict()

    for shape in shapes:
        tinyobj_all_indices = shape.mesh.numpy_indices().reshape(-1, 3)[:, 0]
        if is_mixed_arity and not triangulate:
            raise ArityException(
                "OBJ Loader does not support mixed arities with triangulate=False"
            )
        elif is_mixed_arity:
            these_vertices_per_face = shape.mesh.numpy_num_face_vertices()
            these_faces = np.zeros((0, 3), dtype=index_dtype)
            for this_face in unstack(tinyobj_all_indices, these_vertices_per_face):
                if len(this_face) == 3:
                    these_faces = np.concatenate([these_faces, this_face.reshape(1, 3)])
                else:
                    these_faces = np.concatenate(
                        [
                            these_faces,
                            this_face[[0, 1, 2]].reshape(1, 3),
                            this_face[[0, 2, 3]].reshape(1, 3),
                        ]
                    )
        else:
            these_faces = tinyobj_all_indices.reshape(-1, first_arity)
            if triangulate and first_arity == 4:
                # Triangulate ABCD as ABC + ACD.
                these_faces = these_faces[:, [[0, 1, 2], [0, 2, 3]]].reshape(-1, 3)

        start = len(all_faces)
        end = start + len(these_faces)
        these_face_indices = list(range(start, end))

        all_faces = np.concatenate([all_faces, these_faces])

        for name in shape.name.split():
            if name not in segm:
                segm[name] = []
            segm[name] = segm[name] + these_face_indices

    group_map = GroupMap.from_dict(segm, len(all_faces))
    return Mesh(v=tinyobj_vertices, f=all_faces, face_groups=group_map)


def load(mesh_path, triangulate=False):
    """
    Load a `Mesh` from a path to an OBJ file.

    Args:
        mesh_path (str): A path to an OBJ file
        triangulate (bool): A flag that indicates whether to triangulate the mesh on load.

    Returns:
        lacecore.Mesh: A `Mesh` instance
    """
    reader, config = create_reader_and_config()
    success = reader.ParseFromFile(mesh_path, config)
    if not success:
        raise LoadException(reader.Warning() or reader.Error())
    return _finalize(reader=reader, triangulate=triangulate)


def loads(mesh_string, triangulate=False):
    """
    Load `Mesh` contents from a string.

    Args:
        mesh_string (str): The contents of an OBJ file.
        triangulate (bool): A flag that indicates whether to triangulate the mesh on load.

    Returns:
        lacecore.Mesh: A `Mesh` instance
    """
    reader, config = create_reader_and_config()
    success = reader.ParseFromString(mesh_string, "", config)
    if not success:
        raise LoadException(reader.Warning() or reader.Error())
    return _finalize(reader=reader, triangulate=triangulate)



import numpy as np


def check_arity(faces):
    if faces.shape[1] not in [3, 4]:
        raise ValueError("Expected 3 or 4 vertices per face")


def check_indices(indices, num_elements, name):
    if np.any(indices >= num_elements):
        raise ValueError(
            "Expected indices in {} to be less than {}".format(name, num_elements)
        )



import numpy as np
import vg


def indices_of_original_elements_after_applying_mask(mask):
    """
    Given a mask that represents which of the original elements should be kept,
    produce an array containing the new indices of the original elements. Returns
    -1 as the index of the removed elements.
    """
    result = np.repeat(np.int(-1), len(mask))
    result[mask] = np.arange(np.count_nonzero(mask))
    return result


def create_submesh(
    mesh, vertex_mask, face_mask, ret_indices_of_original_faces_and_vertices=False
):
    """
    Apply the requested mask to the vertices and faces to create a submesh,
    discarding the face groups.
    """
    from .._mesh import Mesh

    new_v = mesh.v[vertex_mask]
    indices_of_original_vertices = indices_of_original_elements_after_applying_mask(
        vertex_mask
    )
    new_f = indices_of_original_vertices[mesh.f[face_mask]]
    submesh = Mesh(v=new_v, f=new_f)

    if ret_indices_of_original_faces_and_vertices:
        indices_of_original_faces = indices_of_original_elements_after_applying_mask(
            face_mask
        )
        return submesh, indices_of_original_faces, indices_of_original_vertices
    else:
        return submesh


def reindex_vertices(mesh, ordering):
    """
    Reorder the vertices of the given mesh, returning a new mesh.

    Args:
        mesh (lacecore.Mesh): The mesh on which to operate.
        ordering (np.arraylike): An array specifying the order in which
            the original vertices should be arranged.

    Returns:
        lacecore.Mesh: The reindexed mesh.
    """
    from .._mesh import Mesh

    vg.shape.check(locals(), "ordering", (mesh.num_v,))
    unique_values, inverse = np.unique(ordering, return_index=True)
    if not np.array_equal(unique_values, np.arange(mesh.num_v)):
        raise ValueError(
            "Expected new vertex indices to be unique, and range from 0 to {}".format(
                mesh.num_v - 1
            )
        )

    return Mesh(v=mesh.v[ordering], f=inverse[mesh.f], face_groups=mesh.face_groups)


def reindex_faces(mesh, ordering):
    """
    Reorder the faces of the given mesh, returning a new mesh.

    Args:
        mesh (lacecore.Mesh): The mesh on which to operate.
        ordering (np.arraylike): An array specifying the order in which
            the original faces should be arranged.

    Returns:
        lacecore.Mesh: The reindexed mesh.
    """
    from .._mesh import Mesh

    vg.shape.check(locals(), "ordering", (mesh.num_f,))
    unique_values = np.unique(ordering)
    if not np.array_equal(unique_values, np.arange(mesh.num_f)):
        raise ValueError(
            "Expected new face indices to be unique, and range from 0 to {}".format(
                mesh.num_f - 1
            )
        )

    return Mesh(
        v=mesh.v,
        f=mesh.f[ordering],
        face_groups=None
        if mesh.face_groups is None
        else mesh.face_groups.reindexed(ordering),
    )



import vg
from ._analysis.analysis_mixin import AnalysisMixin
from ._common.validation import check_arity, check_indices
from ._obj.writer import write as write_obj
from ._selection.selection_mixin import SelectionMixin
from ._transform.transform_mixin import TransformMixin


class Mesh(AnalysisMixin, SelectionMixin, TransformMixin):
    """
    A triangular or quad mesh. Vertices and faces are represented using NumPy
    arrays. Instances are read-only, at least for now. This class is optimized
    for cloud computation.

    Args:
        v (np.ndarray): A `kx3` array of vertices. It will be marked read-only.
        f (np.ndarray): A `kx3` or `kx4` array of vertex indices which make
            up the faces. It will be marked read-only.
        copy_v (bool): When `True`, the input vertices will be copied before
            they are marked read-only.
        copy_f (bool): When `True`, the input faces will be copied before
            they are marked read-only.
    """

    def __init__(self, v, f, copy_v=False, copy_f=False, face_groups=None):
        num_vertices = vg.shape.check(locals(), "v", (-1, 3))
        vg.shape.check(locals(), "f", (-1, -1))
        check_arity(f)
        check_indices(f, num_vertices, "f")

        # TODO: Needs coverage.
        # if copy_f:
        #     f = np.copy(f)
        # if copy_v:
        #     v = np.copy(v)
        f.setflags(write=False)
        v.setflags(write=False)
        self.f = f
        self.v = v
        self.face_groups = face_groups

    # TODO: Needs coverage.
    # @classmethod
    # def from_lace(cls, mesh):
    #     return cls(v=mesh.v, f=mesh.f)

    # TODO: Needs coverage.
    # @classmethod
    # def from_trimesh(cls, mesh):
    #     return cls(v=mesh.vertices, f=mesh.faces)

    def __repr__(self):
        return f"lacecore.Mesh(num_v={self.num_v}, num_f={self.num_f})"

    @property
    def num_v(self):
        """
        The number of vertices.

        Return:
            int: The number of vertices.
        """
        return len(self.v)

    @property
    def num_f(self):
        """
        The number of faces.

        Return:
            int: The number of faces.
        """
        return len(self.f)

    @property
    def is_tri(self):
        """
        `True` if a triangular mesh.

        Return:
            bool: `True` if a triangular mesh.
        """
        return self.f.shape[1] == 3

    @property
    def is_quad(self):
        """
        `True` if a triangular mesh.

        Return:
            bool: `True` if a triangular mesh.
        """
        return self.f.shape[1] == 4

    def write_obj(self, filename):
        """
        Save a mesh's faces, vertices, and face groups to a Wavefront OBJ file.

        Args:
            filename (str): The file to write. If it exists, it will be
                overwritten.
        """
        with open(filename, "w") as f:
            write_obj(f, self)



from ._common.reindexing import reindex_faces, reindex_vertices  # noqa: F401
from ._group_map import GroupMap  # noqa: F401
from ._mesh import Mesh  # noqa: F401
from ._obj.loader import (  # noqa: F401
    ArityException,
    LoadException,
    load as load_obj,
    loads as load_obj_string,
)
from ._obj.writer import write as write_obj  # noqa: F401
from ._selection.selection_object import Selection  # noqa: F401
from ._transform.transform_object import Transform  # noqa: F401
from .package_version import __version__  # noqa: F401



from polliwog import shapes
from .._mesh import Mesh


__all__ = [
    "rectangular_prism",
    "cube",
    "triangular_prism",
]


def _mesh_from_shape_fn(shape_factory_fn, *args, **kwargs):
    vertices, faces = shape_factory_fn(
        *args, ret_unique_vertices_and_faces=True, **kwargs
    )
    return Mesh(v=vertices, f=faces)


def rectangular_prism(origin, size):
    """
    Tesselate an axis-aligned rectangular prism. One vertex is `origin`. The
    diametrically opposite vertex is `origin + size`.

    Args:
        origin (np.ndarray): A 3D point vector containing the point on the
            prism with the minimum x, y, and z coords.
        size (np.ndarray): A 3D vector specifying the prism's length, width,
            and height, which should be positive.

    Returns:
        lacecore.Mesh: A `Mesh` instance containing the rectangular prism.
    """
    return _mesh_from_shape_fn(shapes.rectangular_prism, origin=origin, size=size)


def cube(origin, size):
    """
    Tesselate an axis-aligned cube. One vertex is `origin`. The diametrically
    opposite vertex is `size` units along `+x`, `+y`, and `+z`.

    Args:
        origin (np.ndarray): A 3D point vector containing the point on the
            prism with the minimum x, y, and z coords.
        size (float): The length, width, and height of the cube, which should
            be positive.

    Returns:
        lacecore.Mesh: A `Mesh` instance containing the cube.
    """
    return _mesh_from_shape_fn(shapes.cube, origin=origin, size=size)


def triangular_prism(p1, p2, p3, height):
    """
    Tesselate a triangular prism whose base is the triangle `p1`, `p2`, `p3`.
    If the vertices are oriented in a counterclockwise direction, the prism
    extends from behind them.

    Args:
        p1 (np.ndarray): A 3D point on the base of the prism.
        p2 (np.ndarray): A 3D point on the base of the prism.
        p3 (np.ndarray): A 3D point on the base of the prism.
        height (float): The height of the prism, which should be positive.

    Returns:
        lacecore.Mesh: A `Mesh` instance containing the triangular prism.
    """
    return _mesh_from_shape_fn(
        shapes.triangular_prism, p1=p1, p2=p2, p3=p3, height=height
    )



import setuptools.command.build_py
from setuptools import find_packages, setup

# Set version_info[__version__], while avoiding importing numpy, in case numpy
# and vg are being installed concurrently.
# https://packaging.python.org/guides/single-sourcing-package-version/
version_info = {}
exec(open("lacecore/package_version.py").read(), version_info)

with open("README.md") as f:
    readme = f.read()

with open("requirements.txt") as f:
    install_requires = f.read()

with open("requirements_obj.txt") as f:
    obj_requires = f.read()

exclude = ["**/test_*.py"]


class build_py(setuptools.command.build_py.build_py):
    def find_package_modules(self, package, package_dir):
        import fnmatch

        return [
            (pkg, mod, file)
            for (pkg, mod, file) in super().find_package_modules(package, package_dir)
            if not any(fnmatch.fnmatchcase(file, pat=pattern) for pattern in exclude)
        ]


setup(
    name="lacecore",
    version=version_info["__version__"],
    description="Polygonal mesh library optimized for cloud computation.",
    long_description=readme,
    long_description_content_type="text/markdown",
    author="Metabolize, Minnow Software, Body Labs, and other contributors",
    author_email="github@paulmelnikow.com",
    url="https://github.com/lace/lacecore",
    project_urls={
        "Issue Tracker": "https://github.com/lace/lacecore/issues",
        "Documentation": "https://lacecore.readthedocs.io/en/stable/",
    },
    packages=find_packages(),
    install_requires=install_requires,
    extras_require={"obj": obj_requires},
    cmdclass={"build_py": build_py},
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "Intended Audience :: Manufacturing",
        "Topic :: Artistic Software",
        "Topic :: Multimedia :: Graphics :: 3D Modeling",
        "Topic :: Scientific/Engineering :: Mathematics",
        "Topic :: Scientific/Engineering :: Visualization",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
    ],
)



# Adapted from
# https://github.com/lace/lace/blob/d3c191dffaeedc14aafa4af031d74743de9e632d/lace/serialization/obj/__init__.py
import numpy as np
from .._group_map import GroupMap


def write(fp, mesh):
    """
    Save a mesh's faces, vertices, and face groups to a Wavefront OBJ file.

    Args:
        fp: An open file pointer.
        mesh (lacecore.Mesh): The mesh to write.
    """
    for vertex in mesh.v:
        fp.write("v {} {} {}\n".format(*vertex))

    face_groups = mesh.face_groups or GroupMap.from_dict({}, len(mesh.f))
    last_group_mask = np.zeros(len(face_groups.keys()), dtype=np.bool)
    for i, face in enumerate(mesh.f):
        this_group_mask = face_groups.mask_for_element(i)
        if np.any(last_group_mask != this_group_mask):
            fp.write(
                "g {}\n".format(
                    " ".join(face_groups.group_names_for_element_mask(this_group_mask))
                )
            )
            last_group_mask = this_group_mask

        # Add one, because OBJ indexing is one-based.
        if mesh.is_quad:
            fp.write("f {} {} {} {}\n".format(*face + 1))
        else:
            fp.write("f {} {} {}\n".format(*face + 1))
